\ asm.f     - isforth assembler extension
\ ------------------------------------------------------------------------

 .( loading asm.f ) cr

\s work in progress - will be ready in a year or two maybe :/

\ ------------------------------------------------------------------------

 vocabulary assembler assembler definitions

\ ------------------------------------------------------------------------

 0 var defered              \ address of word to complete last instruction
 0 var opsp                 \ depth on entry into defered opcode

\ ------------------------------------------------------------------------
\ totally headerless : definition

\ does not have a header nor a call to nest (no cfa on this)

here                        \ address of following data
]                           \ switch into compile mode
  sp@ !> opsp exit          \ compile code to save current stack depth
[                           \ switch back to interpret mode

\ ------------------------------------------------------------------------

\ this word is a delay line for executing mneumonic definitions.  each
\ time a mneumonic is called the previous one is executed.

: defer>        ( --- )     \ all mneumonics start with a defer>
  defered                   \ get return address of previous defered op
  r> !> defered             \ save return address of current op
  literal                   \ address of above headerless definition
  >r                        \ remember stack depth after doing previous op
  dup ?:                    \ if there is a previous defered op...
    >r                      \   return into previous op : def
    drop                    \   else discard
  prefix> ;                 \ resolve prefixes on previous opcode

\ ------------------------------------------------------------------------
\ prefixes applied to current and previous opcode

 create [p1] 5 allot        \ buffers to hold prefixes on current and
 create [p2] 5 allot        \ previous instructions

 [p1] var prefixs           \ prefix buffer for current instruction
 [p1] [p2] xor const pmask  \ mask for switching between buffers

\ ------------------------------------------------------------------------

: 0prefixs                  \ zero out both prefix buffers
  [p1] 5 erase
  [p2] 5 erase ; 0prefixs

\ ------------------------------------------------------------------------
\ create a prefix mneumonic

: prefix        ( c1 --- )
  create c,                 \ create prefix word, compile its value
  does>         ( --- c1 )
    c@                      \ fetch prefix value
    prefixs pmask xor       \ inject into buffer for current instruction
    dup c@ 1+       
    2dup swap c!
    + c! ;

\ ------------------------------------------------------------------------
\ compile prefixes for previous instruction if any, swap buffers

: prefix>
  prefixs c@                \ any prefixs on previous instruction ?
  if
    prefixs count           \ compile them
    here swap dup allot
    cmove 
    prefixs off             \ no prefixes in this buffer now
  then

  prefixs pmask xor         \ switch to other prefix buffer
  !> prefixs ;

\ ------------------------------------------------------------------------

  \360 prefix lock        
  \362 prefix repne  \362 prefix repnz
  \363 prefix rep    \363 prefix repe   \363 prefix repz
  \046 prefix es:         
  \056 prefix cs:         
  \066 prefix ss:         
  \076 prefix ds:         
  \144 prefix fs:         
  \145 prefix gs:
  \146 prefix o:          
  \147 prefix a:

\ ------------------------------------------------------------------------

\ it is suggested that you dont use segment overides unless you know you
\ are running in ring zero - they dont make any sense anywhere else
\
\ the operand override will be injected into your code where appropriate
\ but theres nothing to stop you manually injecting it.  if you arent
\ careful though you could end up with two operand overides on one 
\ instruction - dont manually inject unless you know better than the
\ assembler
\
\ it is very doubtful that the address override will actually be used as
\ this assembler does not support 16 bit addresses - it is included here
\ for the sake of consistency - use it if you think its needed

\ ------------------------------------------------------------------------
\ local label buffers

 create locals 40 allot     \ local label address definitions
 create [locals] 400 allot  \ 10 forward references allowed per local
 create counts 10 allot     \ number of forward refs to each local
 create nears 100 allot     \ 10 references per - are they short or near?

\ ------------------------------------------------------------------------

 0 var tmp                  \ temp store for body address of does> word
 0 var ?near                \ branch displacement 32 bits?
 0 var nonear               \ near allowed?

\ ------------------------------------------------------------------------
\ clear all locall label info

: 0locals 
  locals 40 erase 
  [locals] 400 erase
  counts 100 erase
  nears 100 erase ;  0locals

\ ------------------------------------------------------------------------

: near 
  nonear ?exit              \ ignore near if its not allowed
  on> ?near ;

\ ------------------------------------------------------------------------
\ display local label name after an error

: .local        ( n1 --- )
  last @                    \ display word name where local label is used
  count lexmask type
  ." .L" $30 + emit ;       \ disply local label n1

\ ------------------------------------------------------------------------
\ word to create a local label

: local         ( n1 --- )
  create c,                 \ create local lable, compile its number
  does>
    !> tmp                  \ save address of local number
    defer>                  \ complete previous opcode before...
      tmp c@                \ get local label number
      dup 4* locals + dup @ \ is this local already defined?
      if
        drop .local
        true abort"  Redefined"
      then
      nip here swap ! ;     \ set address of this local label

\ ------------------------------------------------------------------------
\ anyone who needs more than 10 local labels needs to rethink instead

  0 local L0:  1 local L1:  2 local L2:  3 local L3:  4 local L4:
  5 local L5:  6 local L6:  7 local L7:  8 local L8:  9 local L9:

\ ------------------------------------------------------------------------
\ fetch byte sized count from address a1 then increment it

: cincr     ( a1 --- count )
  dup c@ tuck 
  1+ swap c! ;

\ ------------------------------------------------------------------------
\ create and handle reference to a local label

: >local        ( n1 --- target )
  create c,
  does>
    c@ dup 4*               \ has local n1 label been defined yet?
    locals + @ ?dup    
    if
      nip exit              \ if so return its address
    then

\ this is a forward reference

    dup>r                   \ keep n1
    40 * [locals] +         \ get address of fwd ref buffer for this local
    r@ counts + cincr tuck  \ get number of forwards to this local
    4* + here swap !        \ store address of referencing instruction 
    nears + r> 4* +         \ remember if this is near or short
    ?near swap c! ;   

\ ------------------------------------------------------------------------
\ create local labels referencing words

  0 >local L0  1 >local L1  2 >local L2  3 >local L3  4 >local L4
  5 >local L5  6 >local L6  7 >local L7  8 >local L8  9 >local L9

\ ------------------------------------------------------------------------
\ resolve n1 references to local label n2

: (locals>)     ( n1 n2 --- )
  dup>r 4* locals + @       \ get target address for local branch
  ?dup 0=
  if
    ." Undefined Local "
    r> .local abort
  then
  swap
  r@ 10 * nears + swap
  r> 40 * [locals] + swap

  for
    3dup
    @ swap c@
    if
      2+ dup>r
      4+ - r> !
    else
      1+ dup>r
      1+ - r> c!
    then        
    4+ swap 1+ swap
  nxt
  3drop ;

\ ------------------------------------------------------------------------

: locals>
  10
  for
    counts r@ + c@ ?dup
    if
      r@ (locals>)
    then
  loop ;

\ ------------------------------------------------------------------------
\ compile a near branch with specified condition code

: jnear         ( target cc --- )
   \017 c,                  \ prefix branch as having a 32 bit offset
   \200 + c,                \ compile opcode with condition code
   here 4+ - , ;            \ compile delta to target

\ ------------------------------------------------------------------------
\ compile a short  branch with specified condition code

: jshort        ( target cc --- )
   \160 + c,                \ compile opcode with condition code
   here 1+ - c, ;           \ compile 8 bit delta to target

\ ------------------------------------------------------------------------
\ branch creating word

: jcc           ( cc --- )
  create c,                 \ create branching word with condition code
  does>         ( target --- )
    !> tmp                  \ save body of compiled cc
    defer>                  \ complete previous instruction
      tmp c@                \ get compiled condition code
      ?near                 \ was near specified for this branch ?
      ?: jnear jshort
      off> ?near ;

\ --------------------------------------------------------------
\ create the branching words

    \000 jcc jo        
    \001 jcc jno
    \002 jcc jb     \002 jcc jnae       \002 jcc jc
    \003 jcc jae    \003 jcc jnb        \003 jcc jnc
    \004 jcc je     \004 jcc jz
    \005 jcc jne    \005 jcc jnz
    \006 jcc jbe    \006 jcc jna
    \007 jcc ja     \007 jcc jnbe
    \010 jcc js
    \011 jcc jns
    \012 jcc jp     \012 jcc jpe
    \013 jcc jnp    \013 jcc jpo
    \014 jcc jl     \014 jcc jnge
    \015 jcc jge    \015 jcc jnl
    \016 jcc jle    \016 jcc jng
    \017 jcc jg     \017 jcc jnle

\ ------------------------------------------------------------------------

: (>loop)        ( target type --- )
  off> nonear
  \340 + c, here 1+ - c, ;

\ ------------------------------------------------------------------------

: >loop 
  create c,
  does>
    !> tmp
    on> nonear              \ near is not allowed
    defers>
      tmp c@ (>loop) ;

\ ------------------------------------------------------------------------

    0 >loop loopne
    1 >loop loope
    2 >loop loop
    3 >loop jcxz

\ ------------------------------------------------------------------------
\ got alota work to do dont i :P




\ ------------------------------------------------------------------------
\ create a new coded definition

forth definitions

: code
  head"                     \ create coded definitions header
  off> defered              \ no previous instructions to assemble yet
  0prefixs 0locals
  sp@ !> opsp
  assembler ;               \ put assembler voc in context

\ ------------------------------------------------------------------------
\ terminate a coded definition

assembler definitions

: end-code
  assembler previous        \ remove assembler voc from context
  reveal                    \ reveal new coded definition
  defer> ;                  \ complete assembly of last instruction
    
\ ========================================================================
