\ block.f       - isforth block file wordset
\ ------------------------------------------------------------------------

.( loading block.f ) cr

\ ------------------------------------------------------------------------
 
 vocabulary block block definitions

\ ------------------------------------------------------------------------

 0 var #bfiles              \ number of open block files
 0 var bham                 \ address of block handle allocation map

\ ------------------------------------------------------------------------
\ add syscall to sync a file with its memory map

 3 144 syscall <msync>      \ belongs in file.f ?

\ ------------------------------------------------------------------------
\ block file handle structure

<headers

struct: blk-hndl
  1 dfield fd               \ file descriptor for block file
  1 dfield mmap             \ memory map address of block file
  1 dfield size             \ memory map size
  1 bfield width            \ block width
  1 bfield height           \ block height
  1 wfield reserved         \ pad out to 16 bytes
;struct

\ ------------------------------------------------------------------------

 0 var hbuf                 \ address of block handle buffer
 0 var namebuff             \ file name buffer for 256 files
 0 var name-list            \ each open files name offset in buffer
 0 var next-name            \ name buffer tail pointer

\ ------------------------------------------------------------------------
\ abort if cant allocate buffers

: alloc     ( size --- address )
  allocate
  not abort" cannot allocate block buffers" ;

\ ------------------------------------------------------------------------
\ create block file handle bufers etc

: alloc-buffers ( --- )
  [ blk-hndl 8 << ] literal \ enough space for 256 block handles
  alloc !> hbuf             \ grab memory from top of head space
  256 alloc !> bham         \ get block handle allocation map
  1024 alloc !> name-list   \ addresses of all open files file names
  40960 alloc dup           \ is 40k enough for up to 256 file names? 
  !> namebuff               \ *** note: - change above buffer to 2k when
  !> next-name ;            \             resize is implemented!!

\ ------------------------------------------------------------------------
\ get address of specified files file name pointer

headers>

: ?name-list     ( handle --- a1 )
  1- 4* name-list + ;

<headers

\ ------------------------------------------------------------------------
\ store file name in file name buffer

: addname       ( name handle --- )
  ?name-list                \ get buffer address for this filename
  next-name dup>r swap !    \ store address of file name in list
  count -1 /string          \ copy filename (with count) to the address
  r> swap dup>r cmove
  r> +!> next-name ;        \ set address for next file name

\ ------------------------------------------------------------------------

 0 var delsize              \ length of name to delete
 0 var deladdr              \ address of name to delete
 0 var s                    \ total length of data to move down in buffer

\ ------------------------------------------------------------------------
\ scan for a non zero item in the name list

: scan-name         ( a1 --- a2 )
  begin
    dup @ ?exit             \ does a1 point to a file name string ?
    4+                      \ no - advance a1
  again ;                   \ above is guaranteed to hit a non zero item

\ ------------------------------------------------------------------------
\ adjust items in name list to account for deleted item

: (delname)         ( a1 --- a2 )
  scan-name dup @           \ scan name list for next entry
  deladdr < ?exit           \ point to a filename above deleted name?

  dup @ c@                  \ get length of string to be moved down
  1+ +!> s                  \ total all bytes to be moved
  delsize over +! ;         \ move reference back by deleted names length

\ ------------------------------------------------------------------------
\ delete file name from file name buffer and move others down to account

: delname       ( handle --- )
  ?name-list dup @          \ get address of file name
  swap off                  \ delete name from list
  decr> #bfiles             \ decrement total number open block files

  #bfiles 0=                \ no others open?
  if                        \ next filename goes to start of buffer
    namebuff !> next-name
    drop exit               \ clean up and exit
  then

  count -1 /string          \ get name len/address including count byte
  negate !> delsize         \ length of name as negative number
  !> deladdr                \ address of name in buffer
  delsize +!> next-name     \ move next name pointer back in buffer

  off> s                    \ number of characters above deleted name

  name-list #bfiles         \ search list for entries above deleted name
  for                       \ and adjust their name list entries
    (delname) 4+
  nxt
  drop s 0= ?exit           \ no filenames above deleted name ?

  deladdr dup delsize -     \ get source and destination for move
  swap s cmove ;            \ get size and move all names down in memory

\ ------------------------------------------------------------------------
\ allocate an unused block handle

: alloc-bh      ( --- handle )
  hbuf 0=                   \ have we grabbed memory for handles ?
  if
    alloc-buffers           \ no, allocate it then
  then

  bham 256 0 scan           \ find unallocated block handle
  1 rot c!                  \ allocate handle
  256 swap - 1+ ;           \ return allocated handle number (1 based)

\ ------------------------------------------------------------------------
\ deallocate block handle

: free-handle   ( handle --- )
  1- bham +                 \ get address of handles allocation map flag
  0 swap c! ;               \ zero it 

\ ------------------------------------------------------------------------
\ calculate address of block handle structure

\ this code assumes the handle structure size is 16 bytes. if this changes
\ then change the code from "4 <<" to "blk-hndl *"

: >handle       ( handle --- handle-structure )
  4 << hbuf + ;
 
\ ------------------------------------------------------------------------
\ close block file and unmap it

headers>

: bclose        ( handle --- )
  dup dup delname           \ delete this files name from name buffer
  >handle dup>r             \ get address of handle structure
  size @ r@ mmap @          \ unmap the block file
  <munmap> drop             \ the unmap worked :P
  r> fd @ fclose            \ now close the file
  free-handle ;             \ and deallocate the files handle

\ ------------------------------------------------------------------------
\ open a block file and memory map it ( always +rw )

: bopen         ( --- handle )
  bl word hhere             \ parse for file name
  2 over fopen ?dup -1 =    \ open file +rw
  if
    hhere count type
    abort"  Open Error"
  then
  dup alloc-bh dup>r        \ allocate a block handle
  >handle dup>r fd !        \ store file descriptor in handle buffer
  64 r@ width c!            \ assume block conforms to default block size
  16 r@ height c!
  1 3 rot fmmap             \ map shared, prot +rw
  r@ size !                 \ store size of mapping in structure
  r> mmap !                 \ and address file got mapped to
  r>                        \ return handle
  tuck addname              \ add name to name list buffer
  incr> #bfiles             \ one more open block file
  dup !> bfh ;              \ make this the default block file

\ ------------------------------------------------------------------------
\ set block dimensions for a given block file

: blksize!      ( width height handle --- )
  >handle tuck              \ get handle structure address
  height c! width c! ;      \ set dimensions

\ ------------------------------------------------------------------------
\ get block dimensions

: blksize@      ( handle --- width height )
  >handle 
  dup width c@
  swap height c@ ;

\ ------------------------------------------------------------------------
\ calculate highest block number 

: bmax      ( handle --- )
  dup >handle size @
  swap blksize@ */ ;

\ ------------------------------------------------------------------------
\ get address of block for given file

: >block        ( block# handle --- a1 )
  swap 1- swap              \ 0 base the block number
  dup blksize@ *            \ get requested block times size of one block 
  rot *
  swap >handle              \ get address of memory map for this file
  mmap @ + ;                \ return address of specified block

\ ------------------------------------------------------------------------
\ get address of given block and address of files handle structure

: block&handle  ( block handle --- block-addr handle-addr )
  tuck >block 
  swap >handle ;

\ ------------------------------------------------------------------------
\ write block back to file (sync)

: bsync         ( block handle --- )
  block&handle
  dup width @
  swap height @ *
  4 -rot swap               \ ms_sync (block till block written :)
  <msync> drop ;            \ assume ok

\ ------------------------------------------------------------------------
\ create new block file - file is not open after creation

: bcreate       ( width height count --- )
  bl word hhere             \ parse file name
  >r * * dup>r              \ calculate total size of block file
  allocate
  dup r@ blank              \ fill entire block file with spaces
  r> swap
  \644 r> fcreate           \ rw-r--r--
  3dup fwrite .             \ write the file
  fclose drop free ;        \ free memory and close file

\ ------------------------------------------------------------------------

 behead forth definitions

\ ========================================================================
