\ inline.f      - isforth macro colon definition creation.
\ ------------------------------------------------------------------------

.( loading inline.f ) cr

\ ------------------------------------------------------------------------

\ an example of what i consider to be some realy realy bad ans style forth
\
\ : foo
\     postpone this postpone that 
\     postpone the-other ; immediate
\
\
\ whats bad here is all the 'postpone' crap which adds huge ammounts of
\ bullshit visual clutter.
\
\ the purpose of all this postponing is to have foo compile stuff inline.
\ everything thats postponed above is compiled into the definition 
\ currently being created. 
\
\ the definition for foo could be : foo this that the-other ; and we 
\ would be compiling a call to foo wherever it was invoked but sometimes
\ when speed is required it is advantageous to compile code inline.
\
\ there are however some limitations to what you can do using the above
\ cluttered method.  you cant very easilly have foo above compile a loop 
\ or a branch or a ." blah" etc inside the target.
\
\ this file gives you a way to dispense with the visuall clutter and with
\ the restrictions.
\
\ these words will create a colon definition that will inject its own code
\ into the definition currently being defined and the macro itself does
\ not take up any space on in the target.
\
\ e.g.

\ m: foo this that the-other ;m
\ : bar .... foo .... ;
\
\ m: blah if ." true" else ." false" then 100 0 do i . loop ;m
\
\ : fud ..... blah ..... ;

\ ------------------------------------------------------------------------

<headers

vocabulary i-voc            \ a headerless vocabulary to create macros in

0 var i-buf                 \ buffer to compile macros to

0 var i-hhere               \ inline header pointer
0 var i-here                \ inline list pointer

0 var x-hhere               \ real dictionary / head space pointers saved
0 var x-here
0 var x-current             \ real current vocabulary

\ ------------------------------------------------------------------------
\ executed at forth init time

\ when forth first runs the i-buf wont have been allocated but it might 
\ have been in some previous incantation of forth (i.e. you did an fsave)
\ make sure forth doesnt think i-buff is allocated till it realy is

: i-0
  defers default
  off> i-buf ;

\ ------------------------------------------------------------------------
\ initialize for compiling into inline buffer 

: >inline
  i-buf 0=                  \ is buffer allocated ?

  if
    16384 allocate drop     \ 8k each for macro definitions and headers
    dup !> i-buf
    dup !> i-here
    8192 + !> i-hhere
  then

  hhere !> x-hhere          \ save here and hhere
  here !> x-here
  current @ !> x-current

  i-hhere !> hp
  i-here !> dp
  i-voc definitions ;       \ compile to inline vocabulary

\ ------------------------------------------------------------------------
\ switch back to compiling into real space

: <inline
  hhere !> i-hhere          \ remember where macro compilation ended
  here !> i-here
  x-hhere !> hp             \ reset hp and dp to their real addresses
  x-here !> dp
  x-current current ! ;

\ ------------------------------------------------------------------------
\ discard all macro code and headers and zero inline vocabulary

headers>

: purge-macros
  [ ' i-voc >body ] literal \ erase all threads in i-voc
  256 erase 
  i-buf dup !> i-here       \ reset macro compilatin pointers
  8192 + !> i-hhere 
  i-voc previous ;          \ remove inline voc from context

\ ------------------------------------------------------------------------

<headers

0 var was-branch            \ true if last token was a branch
0 var m-start               \ start address of current macro definition
0 var m-new                 \ address macro is being inlined to
0 var m-exit                \ exit point of macro

\ ------------------------------------------------------------------------
\ is current token a (.") or a (abort")

: is-quote?     ( a1 token --- a1 token false | a2 true )
  ['] (.") over =           \ is current token a quoted string ?
  over ['] (abort") = or
  if
    ,                       \ yes - compile the token
    4+ count -1 /string     \ get string length and address
    2dup
    here swap               \ allocate space for string
    dup allot cmove         \ copy string into destination
    + true exit             \ indicate to caller that token has been
  then                      \  dealt with
  false ;                   \ else pass token back to caller

\ ------------------------------------------------------------------------
\ words that have a branch vector compiled after them

\ it is assumed that a branch vector is absolute, not relative
\ - this is a true assumption in isforth -

create branches
]
  (nxt)   (do)    (?do)  doif     doelse (loop) 
  (+loop) dobegin ?while dorepeat doagain ?until
[

\ ------------------------------------------------------------------------
\ is the current token a branching type word?

: is-branch?    ( token --- f1 )
  branches 12 pluck 
  dscan nip 0= not
  !> was-branch ;

\ ------------------------------------------------------------------------
\ expand a token from a macro into its target definition

: (m:)      ( a1 token --- a2 )
  is-quote? not             \ if current definition isnt a ." or abort"
  if
    was-branch              \ was the previous token a branch?
    if
      off> was-branch       \ clear flag
      m-start - m-new +     \ relocate branch vector 
    else
      is-branch?            \ is the current token a branch?
    then
    , 4+                    \ compile current token, advance address 
  then ;
    
\ ------------------------------------------------------------------------
\ start a macro colon definition

\ this could be called inline: too

headers>

: m:
  >inline                   \ initialize macro compilation
  create immediate          \ create new word and make it immediate
  >mark                     \ compile dummy macro exit point
  ]                         \ switch into compile mode
  does>                     \ when invoked the macro word will....
    dup @ !> m-exit         \ fetch the compiled exit point of macro
    4+ dup !> m-start       \ point to body of macro
    here !> m-new           \ fetch address to inline macro to

    begin
      dup m-exit <>         \ reached end of macro?
    while
      dup @ (m:)            \ no - fetch and process next token
    repeat
    drop ;                  \ yes - clean up

\ ------------------------------------------------------------------------
\ complete definition of a macro colon definition

\ ;inline ??

: ;m
  >resolve                  \ compile exit point of macro
  [compile] [               \ switch out of compile mode
  <inline ; immediate

\ there is no exit compiled onto the end of a macro : definition

\ ------------------------------------------------------------------------

 behead

\ ========================================================================
