\ tail.f        - isforth command line tail processing
\ ------------------------------------------------------------------------

.( loading tail.f ) cr

\ ------------------------------------------------------------------------

<headers

\ this code should not be included in any turnkey applications. this is
\ the isforth development environment's default arg processing code.
\
\ see args.f for details on how to produce something similar for your own
\ code
 
\ ------------------------------------------------------------------------

 0 var shebang              \ true if we were launched by a shebang
 0 var floading             \ defer fload till after default init

\ ------------------------------------------------------------------------
\ for shebanged forth script files

headers>

: #! 
  floading                  \ dont allow -fload in the #! line
  if
    ." Do not use -fload in the shebang line" cr
    ." Your shebang should use -sfload instead " cr cr
    0 <exit>
  then   
  on> shebang               \ shebang line must contain a -fload 
  [compile] \ ;
    
<headers

\ ------------------------------------------------------------------------
\ exit now if isforth was executed via a shebanged forth source

: ?shebang
  shebang not ?exit
  errno <exit> ;

\ ------------------------------------------------------------------------

: arg-missing?
  arg# argc = 
  if
    cr
    ." Missing Argument"
    cr 0 <exit>
  then ;

\ ------------------------------------------------------------------------

: arg-help
  cr
  ."  -fload FILE               Interpret specified file" cr
  ."  #! isforth -sfload FILE   Place at top of shebanged script" cr
  ."  -help                     Your reading it" bye ;

\ do not use -fload on the shebang line in a script as this will cause the
\ default init chain to run before the script is executed.

\ ------------------------------------------------------------------------
\ list of known args 

create args 
  ," -sfload"               \ fload a shebanged script 
  ," -fload"                \ fload a file specified on the arg list
  ," -help"                 \ display info on args

\ ------------------------------------------------------------------------

: arg>tib
  #tib @ >r                 \ get current length of tib
  arg@ dup strlen           \ get filename string 
  dup #tib +!               \ copy filename into tib
  tib r> + swap cmove ;

\ ------------------------------------------------------------------------
\ execute an fload of specified file

here
  ," fload "

: do-sfload
  arg-missing?              \ fload expects a file name
  literal count dup #tib !  \ copy "fload " to tib
  tib swap cmove
  arg>tib interpret ;       \ interpret specified file

\ ------------------------------------------------------------------------

: do-fload
  arg# !> floading          \ remember current arg position
  argc !> arg# ;            \ halt processing of args till after default

\ ------------------------------------------------------------------------

: arg-init      ( --- )
  off> arg#                 \ number of next arg to process
  args !> arglist           \ address of options table
  4 !> argscount ;          \ number of entries in options table

\ ------------------------------------------------------------------------

: (doargs)
  off> shebang              \ assume not running from #! script
  arg-init                  \ init for arg scan
  begin
    #tib off ?arg           \ is next arg in list known to us?
    >in off
    case:
      0 opt arg-help        \ unknown arg
      1 opt do-sfload       \ fload a shebanged script
      2 opt do-fload        \ fload specified file
      3 opt arg-help        \ display useage info
    ;case
    arg# argc =
  until ;

\ ------------------------------------------------------------------------
\ this word patches itself into the head of the default init chain

: doargs          ( ---- )
  argc                      \ dont try interpret null args
  if 
    (doargs)                \ process args 
    ?shebang                \ quit now if we just ran a #! script
  then                      \ otherwise....
  defers default ;          \ run rest of isforths deafault init chain

\ ------------------------------------------------------------------------
\ this word patches itself onto the end of the default init chain

\ -sfload will be handled prior to any initialization via default so 
\ .hello and .status etc are not dumped to the display for script files.  
\ also, when the script completes forth quits and init never gets run at 
\ all.  this means that scripts cannot use some things that are not 
\ initialized (like the text windowing stuff).
\
\ -fload just sets a flag which tells the following word to do the fload.
\ this word is not executed until everything else in the default init
\ chain has run so everything will have been initialized.  

: do-floading
  defers default
  floading ?dup
  if
    !> arg#
    do-sfload
  then ;

\ ------------------------------------------------------------------------

behead

\ ========================================================================
