#isforth.asm    - isforth main kernel source   (c) 2000+ mark i manning iv
#=========================================================================

 .set ver, 0x010f           #hi byte = maj ver : lo byte = min ver

#-------------------------------------------------------------------------

 .include "macros.1"        #macros to create headers etc

 .section code, "awx", @progbits
 .global _start

#-------------------------------------------------------------------------
#some important variables and constants
 _constant0_ "version", version, ver

 _constant_ "tlist", tlist, 0    #address of top of list space
 _constant_ "thead", thead, 0    #address of top of head space
 _constant_ "endp", endp, _end   #address of end of loaded program space

 _constant_ "arg0", arg0, 0      #isforths own file name
 _constant_ "argp", argp, 0      #pointer to os mangled command line tail
 _constant_ "argc", argc, 0      #number of args passed to isforth
 _constant_ "envp", envp, 0      #pointer to environment

 _constant_ "cols", cols, 0      #width of terminal
 _constant_ "rows", rows, 0      #height of terminal

 _variable_ "'tib", ttib, 0      #address of tib

 _constant_ "shebang", shebang, 0
 _constant_ "intty?", inttyq, 0
 _constant_ "outtty?", outttyq, 0

#-------------------------------------------------------------------------
#headerless pointer to termios structure 

tios:
 bl doconstant
 .int 0

#-------------------------------------------------------------------------
#these need a better home 

 _constant_ "turnkeyd", turnkeyd, 0 #true if we are a turnkeyd app
 _constant_ "lsp", lsp, 0           #fload nest stack pointer
 _variable_ "?inittrm", qinittrm, 0 #flag: term initialized already ?

bhead:      .int 0          #address of bottom of head space

#-------------------------------------------------------------------------
#the beef (moo!)

 .include "reloc.1"         #head space relocation (see fsave.f)
 .include "syscalls.1"      #interface to the BIOS ;)
 .include "stack.1"         #stack manipulation etc
 .include "memory.1"        #fetching and storing etc
 .include "logic.1"         #and/or/xor etc
 .include "math.1"          #basic math functions +/-* etc
 .include "loops.1"         #looping and branching constructs
 .include "exec.1"          #word execution, nest/next etc
 .include "io.1"            #console i/o etc
 .include "compile.1"       #compilation/creating words
 .include "vocabs.1"        #vocabulary creation etc

#-------------------------------------------------------------------------

#expects fd in r3.
_chk_tty:
 liw r4,0x402c7413          #TCGETS ioctl -- see man tcgetattr
 ld_var r5,dp               #here
 li r0,54                   #ioctl syscall #
 sc
 li r3,_true
 bns 0f
 li r3,_false
0:
 blr

#-------------------------------------------------------------------------

chk_tty:
 prolog
 li r3,0                    #stdin
 bl _chk_tty
 st_var r3,inttyq
 li r3,1                    #stdout
 bl _chk_tty
 st_var r3,outttyq
 epilog

#-------------------------------------------------------------------------
#do not define any words below this point unless they are headerless
#-------------------------------------------------------------------------

#-------------------------------------------------------------------------
#allocate forth's list and head space and parameter stack

#isforth's memory runs from 0x10000000 to 0x10100000

init_mem:
 li r0,45                   #sys brk
 lis r3,0x1010              #end of memory...
 sc

#should abort program here if the above fails blah blah

 li r0,125                  #syscall mprotect
 lis r3,0x1000              #start of memory -- see `ld --verbose' or
                            #binutils-xx/ld/emulparams/elf32ppclinux.sh
 lis r4,0x0010              #1MB
 li r5,7                    #rwx permissions
 sc                         #make the entire program space rwx

#should abort program here if the above fails blah blah

 li r0,90                   #sys mmap
 li r3,0                    #void *start    - null
 li r4,0x4000               #size_t length  - 1 page
 li r5,3                    #int prot       - read, write
 li r6,0x0022               #int flags      - anon, priv
 li r7,0                    #int fd         - null
 li r8,0                    #off_t offset   - null
 sc

#should abort program here if the above fails blah blah

 addi rp,r3,0x4000          #set return stack pointer to top of
                            #the buffer we just allocated
 blr

#-------------------------------------------------------------------------
#entry point

_start:                     #stupid linker needs this bullshit
 bl init_mem                #sys brk out to 1MB and sys mprotect to rwx

 li r3,0
 st_var r3,qinittrm         #terminal properties not set yet
 st_var r3,shebang

#isforths memory runs from 0x10000000 to 0x10100000-1

 liw r3,0x100ffff8-36       #absolute top of forth's address space -4
 st_var r3,tios             #termios = 9 cells
 subi r3,r3,120             #enough space to nest 5 floads
 st_var r3,lsp              #don't nest floads!!!
 subi r3,r3,0x100           #256 bytes for terminal input
 st_var r3,ttib

 pop r4                     #get argc
 pop r5                     #get own program name
 st_var r5,arg0
 subi r4,r4,1               #discount arg0
 st_var r4,argc
 cmpwi r4,0
 beq 1f

 mtctr r4
 slwi r4,r4,2               #create argp array
 sub r3,r3,r4
 st_var r3,argp
 mr r4,r3
0:
 pop r5
 stw r5,0(r4)
 addi r4,r4,4
 bdnz 0b

1:
 addi sp,sp,4               #drop blank cell between args and env
 st_var r1,envp             #point to environment

# --

 st_var r3,thead            #top of head space
 ld_var r4,endp             #point to where headers got loaded in
 sub r5,r3,r4               #find distance from _end to top of
                            #unallocated memory ([ttib+v_data] -1)
 srwi r5,r5,1               #divide this distance in 2
 sub r3,r3,r5               #point eax at the mid-point
 st_var r3,tlist            #address of top of list space

 st_var r3,hp               #address for headers to be relocated to
 stwi r3,scr,bhead          #needed by fsave - bottom of head space
 st_var r4,dp               #set dictionary pointer (here)

#--

 st_var rp,rp0              #set address of bottom of return stack
 st_var sp,sp0              #set address of bottom of parameter stack

 bl unpack                  #relocate headers to allocated head space
 bl chk_tty                 #chk if stdin/out are on a terminal

# now we start running FORTH code...

 bl nest                    #the following is a colon definition
 .int default               #initialize forth - run default defered chain
 .int quit                  #run inner function - never returns to here

#-------------------------------------------------------------------------

#in windows you quit by hitting the start button.  in forth you start
#by jumping to quit :)
#
#according to the forth tradition the word quit is the main loop of the
#innter interpreter.  i have a slightly modified view of what quit is
#however - to me quit is the inner compiler.  the interpret loop that 
#(quit) calls is just a means of interfacing to the compiler and testing that 
#which you compile.  
#
#once development is complete and your new application executable is saved
#out the compiling and creating words are usually never referenced again.  
#in a turnkeyd application where all headers have been stripped out of the 
#target executable it would be silly to make a reference to words like 
#(quit) because that path eventually tries to search the dictionary and in 
#a turnkeyd application there is no dictionary.
#
#in a turnkeyd application quit becomes main.  
#
#to make a turnketd application in isforth one would do
#
#  ' app-main is quit
#  turnkey app-filename
#
#when the application is executed the call to quit above will jump to your
#app-main.  if you do not want anything initialized by the isforth kernel 
#(like the terminal etc) you could do...
#
#  ' app-main is default 
#  turnkey app-filename
#
# ...instead.

#-------------------------------------------------------------------------
#marks end of code space (where boot will set dp pointing to)

#note:   do not define anything at all below this point 

_end:                       #when isforth loads, this is where headers are

#=========================================================================
