\ terminfo.f    - isforth terminfo handling words
\ -------------------------------------------------------------------------

 .( loading terminfo.f ) cr

\ ------------------------------------------------------------------------

 terminal definitions

\ ------------------------------------------------------------------------

\ this file defines various words with terminfo escape sequence names.
\ when you invoke one of these the code parses and interprets the correct
\ terminfo format string and compiles an escape sequence which can then be 
\ output
\
\ not all escape sequences are being supported, most we will never need but 
\ should you require one that is missing let me know and i will add it
 
\ ------------------------------------------------------------------------

<headers

create info-file
  ," /usr/share/terminfo/x/" 
  here 10 allot 10 erase    \ space filled in with term name
 
create env_term
  ," TERM"

\ ------------------------------------------------------------------------
 
 0 var terminfo             \ address of terminfo data 
 0 var tsize                \ size of memory mapping

\ we never actually unmap the terminfo file so we dont realy need to 
\ remember the size of the mapping

\ ------------------------------------------------------------------------
\ escape sequences compiled to this buffer

 0 var $buffer              \ output string compile buffer
 0 var #$buffer             \ number of characters compiled to $buffer

\ ------------------------------------------------------------------------
\ pointers to each section within terminfo file

\ these are realy constants but we dont kmow their values yet

 0 var t-names              \ names section
 0 var t-bool               \ bool section
 0 var t-numbers            \ numbers section
 0 var t-strings            \ string section (offsets within following)
 0 var t-table              \ string table

\ ------------------------------------------------------------------------
\ various buffers used when parsing escape sequence format strings

 0 var f$                   \ format string parse address
 0 var fsp                  \ format string stack pointer
 0 var params               \ format string parameters
 0 var a-z                  \ format string variables
 0 var A-Z                  \ format string variables

\ ------------------------------------------------------------------------
\ create full path to terminfo file

\ this code ignores the fact that a user can have a ~/.terminfo directory

: (get-info)    ( a1 n1 ---  )
  over c@                   \ get first character of terminfo file name
  [ info-file $15 + ] 
  literal c!                \ overwrite x
  dup                       \ get copy of length of file name
  info-file c@              \ get length of above path
  +                         \ add the lengths
  info-file c!              \ length of path and file name
  [ info-file $17 + ] literal
  swap cmove ;              \ append name to path

\ ------------------------------------------------------------------------
\ read terminfo file (memory map it)

: read-info
  0 info-file fopen         \ open terminfo file for read
  dup -1 =
  if
    ." Unknown TERM: "
    info-file count type    \ display unknown terminfo file
    bye                     \ and get out
  then

  1 dup pluck fmmap         \ map shared and prot read
  !> tsize !> terminfo
  fclose ;                  \ close the file but keep the mapping

\ ------------------------------------------------------------------------
\ allocate a buffer of n1 bytes in size

: ?alloc        ( n1 --- a1 )
  allocate ?exit
  ." Cannot allocate terminal buffers"
  bye ;

\ ------------------------------------------------------------------------
\ alloate terminal output buffer

: alloc-buffers
  32768 ?alloc !> $buffer   \ sequence output buffer
  20 ?alloc 20 + !> fsp     \ format string stack
  36 ?alloc !> params       \ format string parameter buffer
  104 ?alloc !> a-z         \ format string variable buffers
  104 ?alloc !> A-Z ;

\ ------------------------------------------------------------------------
\ initialize pointers to each section within terminfo file

: init-pointers
  terminfo dup>r 12 + !> t-names
  r@ 2+ lew@ t-names + !> t-bool
  r@ 4+ lew@ t-bool + dup 1 and + !> t-numbers
  r@ 6 + lew@ 2* t-numbers + !> t-strings
  r> 8 + lew@ 2* t-strings + !> t-table ;

\ ------------------------------------------------------------------------

: valid?
  terminfo lew@ \0432 = ?exit
  ." Terminfo - Bad Magic!" 
  bye ;

\ ------------------------------------------------------------------------
\ read terminfo file and calculate addresses for each section therein

: get-info
  env_term getenv           \ did anyone bother to set $term env ?
  if
    (get-info) read-info    \ create full path to terminfo file and read 
    valid?                  \ test sanity of terminfo file
    alloc-buffers           \ allocate buffers
    init-pointers           \ initialize pointers to each section 
  else
    ." terminfo - I'm confused (or you are)"
    bye
  then ;

\ ------------------------------------------------------------------------
\ compile char c1 to output sequence string buffer

headers>

: c>$           ( c1 --- )
  $buffer #$buffer + c!     \ append to $buffer
  incr> #$buffer ;          \ increment compiled string length

<headers

\ ------------------------------------------------------------------------
\ fetch next character of terminfo format string

: f$@          ( --- c1 )
  f$ c@ incr> f$ ;

\ ------------------------------------------------------------------------
\ push item n1 onto format string stack

: >fsp          ( n1 --- )
  -4 +!> fsp                \ pre-decrement stack pointer
  fsp ! ;                   \ store n1 at address pointed to by fsp

\ ------------------------------------------------------------------------
\ pop item n1 from format string  stack

: fsp>          ( --- n1 )
  fsp @                     \ fetch item pointed to by fsp
  4 +!> fsp ;               \ post increment 

\ ------------------------------------------------------------------------
\ pop 2 items off format parameter stack

: 2fsp>         ( --- n1 n2 )
  fsp> fsp> ;

\ ------------------------------------------------------------------------
\ various format string token handlers

: %% '%' c>$ ;
: %c fsp> c>$ ;
: %& 2fsp> and >fsp ;
: %| 2fsp> or >fsp ;
' %& alias %A
' %| alias %O
: %! fsp> 0= >fsp ;
: %~ fsp> not >fsp ;
: %^ 2fsp> xor >fsp ;
: %+ 2fsp> + >fsp ;
: %- 2fsp> swap - >fsp ;
: %* 2fsp> * >fsp ;
: %/ 2fsp> swap / >fsp ;
: %m 2fsp> swap % >fsp ;
: %= 2fsp> = >fsp ;
: %> 2fsp> swap > >fsp ;
: %< 2fsp> swap < >fsp ;
: %' f$@ c>$ f$@ drop ;

\ ------------------------------------------------------------------------
\ increment first 2 parameters for ansi terminals

: %i 
  params dup incr
  4+ incr ;

\ ------------------------------------------------------------------------
\ not too sure about these two

: %s fsp> count bounds do i c@ c>$ loop ; 
: %l fsp> c@ >fsp ;

\ ------------------------------------------------------------------------
\ point to specific variable (static or dynamic)

: ?a-z      ( --- a1 )
  f$@ 'a' 'z' between       \ is next char in format lower case alpha?
  if
    'a' - a-z               \ yes - set dynamic variable
  else
    'A' - A-Z               \ no - set static variable
  then
  swap 4* + ;

\ ------------------------------------------------------------------------
\ fetch and store variables

: %P        ( --- )  fsp> ?a-z ! ;
: %g        ( --- )  ?a-z @ >fsp ;

\ ------------------------------------------------------------------------
\ parse number from format string and write to output string

: %{
  0
  begin
    f$@ dup '}' <>
  while
    swap 10 * +
  repeat
  >fsp ;
  
\ ------------------------------------------------------------------------
\ thesr are both noops

' noop alias %?             \ start a conditional
' noop alias %;             \ end a conditional

\ ------------------------------------------------------------------------
\ this is where we actually test and act on the condition

: %t                        \ we are currently pointing to the true part
  fsp> ?exit                \ if true parse true part
  begin                     \ else skip to end of true part
    begin
      f$@ '%' =
    until
    f$@ 'e' ';' either      \ 'else' and 'endif' can both end an 'if'
  until ;

\ ------------------------------------------------------------------------
\ executing this it means we have just executed a true part

: %e 
  begin                     \ skip past else part and elseif part to endif
    begin
      f$@ '%' =
    until
    ';' =
  until ;

\ ------------------------------------------------------------------------
\ fetch parameter from format string parameter buffer

: %p
  f$@ $f and
  1- 4* params + @
  >fsp ;

\ ------------------------------------------------------------------------

0 var #d                    \ 2 or 3 digits required (see below)

: (%d)      ( n1 --- )
  #d 0=
  if
    0 <# #s #> 
  else
    0 <# #d for # nxt #>
  then ;

\ ------------------------------------------------------------------------
\ write number to output sequence

: %d 
  base @ decimal            \ make sure we're in decimal
  fsp> (%d)                 \ get number to asciify
  dup>r                     \ remember string length
  $buffer #$buffer +        \ point to $buffer current position
  swap cmove                \ append asciified number
  r> +!> #$buffer           \ add string length to $buffer length
  base ! ;                  \ restore base

\ ------------------------------------------------------------------------

: ?digit        ( c1 --- )
  dup '2' '3' either
  if
    $0f and !> #d 
    f$@                     \ this better be a 'd' :)
  then ;

\ ------------------------------------------------------------------------
\ we parsed a % char from format string.

: (%)       ( ... c1 --- )
  f$@ nip                   \ get % command char from format string
  ?digit
  case:                     \ execute command
    '%' opt %%  'p' opt %p
    'd' opt %d  'c' opt %c
    'i' opt %i  '&' opt %&
    '|' opt %|  '^' opt %^
    '+' opt %+  '-' opt %-
    '*' opt %*  '/' opt %/
    'm' opt %m  '=' opt %=
    '>' opt %>  '<' opt %<
    'A' opt %A  'O' opt %O
    $27 opt %'  '{' opt %{
    'P' opt %P  'g' opt %g
    't' opt %t  'e' opt %e
    ';' opt %;
  ;case ;

\ ------------------------------------------------------------------------
\ compile a sequence string to the string buffer

: (format)      ( --- )
  begin
    f$@ ?dup                \ get next char of format string
  while
    '%' over =              \ if its a % (a command)
    ?:
      (%)                   \ ...execute command
      c>$                   \ otherwise add this char to output string
  repeat ;

\ ------------------------------------------------------------------------
\ write compiled escape sequence to stdout

headers>

: (.$buffer)    ( --- )
  #$buffer $buffer          \ count address
  1 <write> drop            \ stdout
  $buffer 32768 erase 
  off> #$buffer ;

\ ------------------------------------------------------------------------

 defer .$buffer 
 ' (.$buffer) is .$buffer

\ ------------------------------------------------------------------------
\ store n1 parameters in paramter table

<headers

: >params       ( ... n1 --- )
  params 36 erase         \ zero all parameters
  ?dup 0= ?exit
  for
    r@ 4* params + !
  nxt ;

\ ------------------------------------------------------------------------
\ create word to handle format string n1 with n2 parameters

\ n1 is the offset within the terminfo string section

: format      ( n1 n2 --- )
  create , ,
  does>
    dcount swap @
    t-strings + lew@ 
    dup $ffff =             \ empty format string?
    if 
      2drop exit
    then
    t-table + !> f$
    >params (format) ;

\ an empty format string is not always an error.  e.g. the linux terminal
\ has an empty enacs because the alt charset does not need enabling

\ ------------------------------------------------------------------------
\ translate alt charset character prior to emit (not done automatically)

headers>

: >acsc      ( c1 --- c2 )
  t-strings 292 + lew@
  t-table + 60 pluck scan
  if
    1+ c@ nip
  else
    drop
  then ;

<headers

\ ------------------------------------------------------------------------
\ a hell of alot of work.... (way too much :)

\ v - offset within terminfo string section
\   v - number of parameters to pass to this word
\                         v - * means supported by eterm but not by me

\ 0 0 format back_tab       \ back tab
  2 0 format (bell)         \ ding!
  4 0 format (cr)           \ carriage return
  6 2 format (csr)          \ change scroll region
\ 8 tbc                   * \ clear all tabs
 10 0 format (clear)        \ clear screen and home cursor
 12 0 format (el)           \ clear to end of line
 14 0 format (ed)           \ clear to end of screen
 16 1 format (hpa)          \ horrizontal position
\ 18 cmdch                  \ terminal settable command character 
 20 2 format (cup)          \ set cursor position
 22 0 format (cud1)         \ cursor down one
 24 0 format (home)         \ cursor home (if no cup)
 26 0 format (civis)        \ cursro invisible
 28 0 format (cub1)         \ cursor back one
\ 30 mrcup                  \ memory relative cursor addressing
 32 0 format (cnorm)        \ cursor normal
 34 0 format (cuf1)         \ cursor forward one
\ 36 ll                     \ cursor to lower left
 38 0 format (cuu1)         \ cursor up one
\ 40 cvvis                  \ make cursor verry visible
 42 0 format (dch1)         \ delete one character
 44 0 format (dl1)          \ delete line
\ 46 dsl                    \ display status line
\ 48 hd                     \ down half line
 50 0 format (smacs)        \ enable alt charset mode
\ 52 0 format (blink)     * \ enter blink mode
  54 0 format (bold)        \ enter bold mode
  56 0 format (smcup)       \ strings to start program using cup
\ 58 smdc                   \ enter delete mode
\ 60 dim                    \ enter dim mode
 62 0 format (smir)         \ enter insert mode
\ 64 invis                  \ enter secure mode
\ 66 prot                   \ enter protected mode
 68 0 format (rev)          \ enter reverse mode
 70 0 format (smso)         \ enter standout mode
 72 0 format (smul)         \ enter underline mode
 74 1 format (ech)          \ erase chars
 76 0 format (rmacs)        \ exit alt charset mode
 78 0 format (sgr0)         \ turn off all attributes
 80 0 format (rmcup)        \ string to end programs using cup (eh?)
\ 82 rmdc                   \ exit delete mode
 84 0 format (rmir)         \ exit insert mode
 86 0 format (rmso)         \ exit standout mode
 88 0 format (rmul)         \ exit underline mode
\ 90 flash                * \ flash screen (evil)
\ 92 ff                     \ form feed
\ 94 fsl                    \ from status line
 96 0 format (is1)          \ init string 1
 98 0 format (is2)          \ init string 2
\ 100 is3                   \ init string 3
\ 102 if                    \ name of init file
104 0 format (ich1)         \ insert char
106 0 format (il1)          \ insert line

\ ip                        \ insert padding after inserted char

\ 202 rmm                   \ meta mode off
\ 204 smm                   \ meta mode on
\ 206 nel                   \ behave like cr followed by lf
\ 208 pad                   \ pad char instead of null
210 1 format (dch)          \ delete characters
\ 212 dl                  * \ delete lines
214 1 format (cud)          \ cursor down lines
216 1 format (ich)          \ insert characters
218 1 format (indn)         \ scroll forward lines
\ 220 il                  * \ insert lines
222 1 format (cub)          \ cursor left chars
224 1 format (cuf)          \ cursor forward chars
226 1 format (rin)          \ scroll back lines
228 1 format (cuu)          \ cursor up lines
\ 230 pfkey                 \ program function key
\ 232 pfloc                 \ program function key
\ 234 pfx                   \ program function key
\ 236 mc0                   \ print screen
\ 238 mc4                 * \ printer off
\ 240 mc5p                * \ printer on
\ 242 rep                   \ repeat char
\ 244 rs1                 * \ reset first string
\ 246 rs2                 * \ reset second string
\ 248 rs3                   \ reset thrid string
\ 250 rf                    \ reset file
252 0 format (rc)           \ restore cursor
254 1 format (vpa)          \ row address
256 0 format (sc)           \ save cursor
258 0 format (ind)          \ scroll forward
260 0 format (ri)           \ scroll backwards
262 9 format (sgr)          \ set attributes
264 0 format (hts)          \ set tab
\ 266 wind                  \ set window
268 0 format (ht)           \ tab
\ 270 tsl                   \ to status line
\ 272 uc                    \ underline char
\ 274 hu                    \ up half line
\ 276 iprog                 \ init prog

\ 288 mc5p                  \ turn on printer for bytes
\ 290 rmp                   \ like ip but when in insert mode
\ 292 acsc      translation table, not an output string
\ 294 pln                   \ program label 

\ 298 smxon                 \ enter xon/xoff mode
\ 300 rmxon                 \ exit xon/xoff mode
302 0 format (smam)         \ turn on auto margins
304 0 format (rmam)         \ turn off auto margins
\ 306 xonc                  \ xon character
\ 308 xoffc                 \ xoff character
310 0 format (enacs)        \ enable alt charset mode
\ 312 smln                  \ turn on soft labels
\ 314 rmln                  \ turn off soft labels
\ 430 rfi                   \ send next input char (for ptys)

538 0 format (el1)          \ clear to beginning of line
\ 540 mgc                   \ clear margins
\ 542 smgl                  \ set left soft margin at current column
\ 544 smgr                  \ set right soft margin at current column
\ 546 fln                   \ label format
\ 548 sclk                  \ set clock hors, mins, secs
\ 550 dclk                  \ display clock
\ 552 rmclk                 \ remove clock
\ 554 cwin                  \ define window 
\ 556 wingo                 \ go to window
\ 558 hup                   \ hang up
\ 560 dial                  \ dial number
\ 562 qdial                 \ dial number without checking
\ 564 tone                  \ select touch tone dialing
\ 566 pulse                 \ select pulse dialing
\ 568 hook                  \ flash switch hook
\ 570 pause                 \ pause for 2-3 seconds
\ 572 wait                  \ wait for dial-tone
\ 574 u0                    \ user string 0
\ 576 u1                    \ user string 1
\ 578 u2                    \ user string 2
\ 580 u3                    \ user string 3
\ 582 u4                    \ user string 4
\ 584 u5                    \ user string 5
\ 586 u6                    \ user string 6
\ 588 u7                    \ user string 7
\ 590 u8                    \ user string 8
\ 592 u9                    \ user string 9
594 0 format (op)           \ set default pair to its original value
\ 596 oc                    \ set all color pairs to the original ones
\ 598 initc                 \ initialize color
\ 600 initp                 \ initialize color pair
\ 602 scp                   \ set current color pair
\ 604 setf                  \ set foreground color
\ 606 setb                  \ set background color
\ 608 cpi                   \ change number of characters per inch
\ 610 lpi                   \ change lines per inch
\ 612 chr                   \ change horizontal resolution
\ 614 cvr                   \ change vertical resolution
\ 616 defc                  \ define character
\ 618 swidm                 \ enter doublewide mode
\ 620 sdrfq                 \ enter draft quality mode
\ 622 sitm                  \ enter italic mode
\ 624 slm                   \ enter leftward mode
\ 626 smicm                 \ enter micro motion mode
\ 628 snlq                  \ enter near letter quality mode
\ 630 snrmq                 \ enter normal quality mode
\ 632 sshm                  \ enter shadow print mode
\ 634 ssubm                 \ enter subscript mode
\ 636 ssupm                 \ enter superscript mode
\ 638 sum                   \ start upward cariage motion
\ 640 swidm                 \ end double wide mode
\ 642 ritm                  \ end italics mode
\ 644 rlm                   \ end left motion mode
\ 646 rmicm                 \ end micro motion mode
\ 648 rshm                  \ end shadow print mode
\ 650 rsubm                 \ end subscript mode
\ 652 rsupm                 \ end superscript mode
\ 654 rum                   \ end reverse character motion
\ 656 mhpa                  \ like column address in micro mode
\ 658 mcud1                 \ like cursor down in micro mode
\ 660 mcub1                 \ like cursor left in micro mode
\ 662 mcuf1                 \ like cursor right in micro mode
\ 664 mvpa                  \ like row address in micro mode
\ 666 mcuu1                 \ like cursor up in micro mode
\ 668 porder                \ match software bits to print head pins
\ 670 mcud                  \ like param down cursor in micro mode
\ 672 mcu1                  \ like param left cursor in micro mode
\ 674 mcuf                  \ like param right cursor in micro mode
\ 676 mcuu                  \ like param cursor up in micro mode
\ 678 scs                   \ select character set
\ 680 smbm                  \ set bottom margin at current line
\ 682 smgbp                 \ set bottom marting at line 
\ 684 smgl                  \ set left soft margin at current column
\ 686 smgrp                 \ set right margin at column
\ 688 smgt                  \ set top margin at current line
\ 690 smgtp                 \ set top (bottom) margin at row
\ 692 sbim                  \ start printing bit image graphics
\ 694 scsd                  \ start character set definition
\ 696 rbim                  \ stop printing bit image graphics
\ 698 rcsd                  \ end definition of character set
\ 700 subcs                 \ list of subscriptabloe characters
\ 702 supcs                 \ list of superscriptable characters
\ 704 docr                  \ printing any of these characters causes cr
\ 706 serom                 \ no motion for subsequent character
\ 708 csnm                  \ produce item from list of char set names
\ 712 minfo                 \ mouse status information
\ 714 reqmp                 \ request mouse position
\ 716 getm                  \ get button events (parametr is undoc)
718 1 format (setaf)        \ set foreground color
720 1 format (setab)        \ set background colorc

\ ------------------------------------------------------------------------
\ create word to execute one of the above and then output the result

: escape 
  create ' , 
  does> 
    @ execute
    .$buffer ;

\ ------------------------------------------------------------------------

headers>

escape bell  (bell)     escape cr    (cr)       escape csr   (csr)
escape clear (clear)    escape el    (el)       escape ed    (ed)
escape hpa   (hpa)      escape cup   (cup)      escape cud1  (cud1)
escape home  (home)     escape civis (civis)    escape cub1  (cub1)
escape cnorm (cnorm)    escape cuf1  (cuf1)     escape cuu1  (cuu1)
escape dch1  (dch1)     escape dl1   (dl1)      escape smacs (smacs)
escape bold  (bold)     escape smcup (smcup)    escape smir  (smir)     
escape rev   (rev)      escape smso  (smso)     escape smul  (smul)     
escape ech   (ech)      escape rmacs (rmacs)    escape sgr0  (sgr0)     
escape rmcup (rmcup)    escape hts   (hts)      escape ht    (ht)       
escape smam  (smam)     escape rmam  (rmam)     escape rmir  (rmir)     
escape rmso  (rmso)     escape rmul  (rmul)     escape is1   (is1)      
escape is2   (is2)      escape ich1  (ich1)     escape il1   (il1)      
escape dch   (dch)      escape cud   (cud)      escape ich   (ich)
escape indn  (indn)     escape cub   (cub)      escape cuf   (cuf)
escape rin   (rin)      escape cuu   (cuu)      escape rc    (rc)
escape vpa   (vpa)      escape sc    (sc)       escape ind   (ind)
escape ri    (ri)       escape sgr   (sgr)      escape hts   (hts)
escape ht    (ht)       escape enacs (enacs)    escape setaf (setaf)
escape setab (setab)

\ ------------------------------------------------------------------------
\ wrappers for some of the above so forth can keep track of the cursor

: clear #out off #line off clear ;
: hpa   dup #out ! hpa ;
: cup   2dup #out ! #line ! cup ;
: cud1  #line incr cud1 ;
: home  1 dup !> #line !> #out home ;
: cub1  dup negate #out +! cub1 ;
: cuf1  #out incr cuf1 ;
: cuu1  #line decr cuu1 ;
: dch1  #out decr dch1 ;
: dch   #out decr dch ;
: cud   dup #line +! cud ;
: ich   #out incr ich ;
: cub   dup negate #out +! cub ;
: cuf   dup #out +! cuf ;
: cuu   dup negate #line +! cuu ;
: vpa   dup #line ! vpa ;
: cr    cr #out off #line @ 1+ rows min #line ! ;

\ ------------------------------------------------------------------------

 0 var acs                  \ flag - is alt charset enabled?

\ ------------------------------------------------------------------------
\ switch to alt charset

: >alt          ( --- ) 
  acs 0=                    \ is alt charset enabled?
  if                        \ we dont want to keep enabling it over and 
    on> acs                 \ over if its already enabled
    enacs
  then
  smacs ;                   \ set mode alt char set

\ ------------------------------------------------------------------------

 ' rmacs alias <alt         \ remove mode alt char set
 ' cup alias at
 ' civis alias curoff
 ' cnorm alias curon

\ ------------------------------------------------------------------------
\ enquire terminal size on program entry 

: ?term
  pad $40087468 0 <ioctl>   \ get window size using ioctl
  drop                      \ drop return value (0 on success)
  pad w@ !> rows            \ set window size variables
  pad 2+ w@ !> cols ;

\ ------------------------------------------------------------------------ 
\ forth will need to know what the terminal's display size is!

<headers

: xyzzy 
  defers default
  get-info ?term ;

headers>

\ ------------------------------------------------------------------------

 forth definitions

\ ========================================================================
