\ screen.f      - isforth text user interface screen handling words
\ ------------------------------------------------------------------------

 .( loading screen.f ) cr

\ ------------------------------------------------------------------------
\ copy all windows attached to screen into screen buffer 1

<headers

: .windows      ( s --- )
  s.window @                \ get address of first window in chain
  begin
    ?dup                    \ reached end of chain?
  while
    dup .window             \ no, draw window into screen
    w.next @                \ link back to previous window in chain
  repeat ;

\ ------------------------------------------------------------------------

 0 var attr                 \ current attributes

\ ------------------------------------------------------------------------
\ set requested foreground/background/bold  attributes

: ?>attrib     ( w1 --- )
  $ffff and dup !> attr     \ remember attributes
  dup $ff and >attrib       \ set fg/bg
  8 u>> dup !> pref >pref   \ set bold/underline
  $10 and ?: >alt <alt ;    \ set alt/charset

\ ------------------------------------------------------------------------
\ position cursor to correct coordinates unless its already there

: s-at          ( x y --- )
  swap 2dup
  #line @ #out @ d=
  ?:
    2drop
    at ;

\ ------------------------------------------------------------------------
\ translation of alt charset characters is not done automatically

: (s-emit)      ( c1 --- )
  pref $10 and
  if
    >acsc
  then
  emit ;

\ ------------------------------------------------------------------------
\ output one character of screen data to console 

: s-emit       ( n1 i --- )
  sw /% s-at                \ position cursor (maybe)
  16 u>> (s-emit)           \ shift out attribs to get char itself
  #out incr ;               \ bug fix :)

\ None of these characters are actually being output to the display yet,
\ they are all being compiled into the $buffer in term.f.  Nothing is
\ written to the display till we have the entire display compiled, so
\ #out isnt being updated by emit - this meant that I was not keeping
\ track of where the cursor was.  So, for every single character on the 
\ display I was setting the cursor position - this was slowing me down
\ a lot and was also subject to some weird error that sometimes misplaced
\ characters.


\ ------------------------------------------------------------------------
\ A character is a 4-byte value.  The bytes are (high to low):
\  unused
\  character
\  flags (bold/underline/etc.)
\  colors (high 4 bits are bg, low 4 are fg)

\ ------------------------------------------------------------------------
\ fetch 4 byte character and attribs from buffer at address a1

: @chr      ( a1 --- ) 
  @ $ffffff and ;

\ ------------------------------------------------------------------------

 0 var c1                   \ current char from screen buffer 1
 0 var s1                   \ current address within screen buffer 1
 0 var s2                   \ current address within screen buffer 2

\ ------------------------------------------------------------------------
\ see if character has changed since last update

: s1s2=             ( index --- f1 )
  4* dup                    \ calculate offset within buffers
  sb1 + dup !> s1           \ calculate address within buffer 1
  @chr dup !> c1            \ get char from buffer 1
  swap sb2 + dup !> s2      \ calculate address within buffer 2
  @chr <> ;                 \ char came in both buffers ?

\ ------------------------------------------------------------------------
\ copy changed character from buffer 1 to buffer 2

: s1>s2             ( --- )
  s1 s2 4 cmove ;

\ ------------------------------------------------------------------------
\ output changed characters of screen to console

: (.screen)                 \ too much if/and/but loop nesting
  sw sh *                   \ calculate char size of window (not byte)
  for                       \ for each char of screen do (end to start)
    r@ s1s2=                \ has character changed since last update?
    if
      s1>s2                 \ yes, update buffer2 of screen
      c1 dup ?>attrib       \ set attribs for this char
      r@ s-emit             \ emit the character

\ Now search for all other characters in the screen with these attribs
\ and output them too.  Moving the cursor all over the place is faster
\ then constantly switching attributes (thanx to kc5 for pointing that
\ out to me).

      r@ 0                  \ scan from start of screen to current pos
      ?do
        i s1s2=             \ did i'th char change?
        c1 $ffff and        \ mask in attribute bytes of this char
        attr = and          \ same as current attribs ?
        if
          s1>s2             \ yes, update buffer 2 of screen
          c1 i s-emit       \ and write char to output buffer
        then
      loop
    then 
  nxt ;

\ ------------------------------------------------------------------------
\ compile huge string of ansi shit for display then blast it all out

headers>

: .screen       ( s --- )
  pref >r                   \ save current terminal attributes
  dup .windows              \ draw all windows into screen buffer 1

  dup s.buffer1 @ !> sb1    \ get address of this screen's first and second
  dup s.buffer2 @ !> sb2    \ buffers
  dup s.width w@ !> sw      \ get this screen's width and height
      s.height w@ !> sh

  ['] noop is .$buffer      \ prevent write of escape sequences till later
  ['] c>$ is emit           \ emit characters into buffer, not to display

  (.screen)                 \ compile new display

  ['] (.$buffer) is .$buffer \ restore these to their defaults
  ['] (emit) is emit 

  .$buffer                  \ and blast entire display in one go

  r> !> pref >pref ;        \ restore terminal attributes

\ ------------------------------------------------------------------------
\ scan to end of window chain

<headers

: (attach)      ( w --- last-window.next )
  begin
    dup w.next @            \ fetch address of next window in list
    ?dup                    \ was there a next window ?
  while
    nip                     \ yes, discard previous window address
  repeat 
  w.next ;                  \ point to empty w.next field - to be filled

\ ------------------------------------------------------------------------
\ attach window w to screen s

headers>

: attach        ( s w --- )
  2dup w.screen !           \ store address of parent screen in window
  dup w.next off            \ no windows linked to this one yet
  swap                      \ w s ---

  s.window dup @ ?dup
  if
    nip (attach)
  then
  ! ;                       \ store w in link address

\ ------------------------------------------------------------------------
\ search linked window list for window to detach

<headers

: (detach)
  sb1 s.window @            \ fetch address of first window in screen
  begin                     \ we know its not the one we are detaching
    dup w.next @            \ fetch next window in chain
    dup wb <>               \ if its not the window to detach
  while
    nip
  repeat
  drop
  wb w.next @               \ get next window after one to delete
  swap w.next ! ;           \ link it to window before one to delete

\ ------------------------------------------------------------------------
\ detach a window from its associated screen

headers>

: detach        ( w --- )
  dup !> wb                 \ remember window structure address
  w.screen @                \ get its associated screen
  dup !> sb1                \ remember screen structure address
  s.window @ wb =           \ is our window the first one attached?
  if
    wb w.next @             \ yes, get address of next window if any
    sb1 s.window !          \ store this as first one in screen
  else
    (detach)                \ no, search window list for one to delete
  then 
  wb w.screen off ;

\ ------------------------------------------------------------------------
\ allocate buffers for screen s

: salloc        ( s --- f )
  dup>r s.width w@ 4*       \ get byte size of one line of screen
  r@ s.height w@ *          \ times height (equals # bytes in screen)
  dup 2* allocate           \ allocate two buffers for screen
  if
    dup r@ s.buffer1 !
    +   r@ s.buffer2 !
    true
  else
    drop false
  then 
  r>drop ;

\ ------------------------------------------------------------------------
\ set width w and height h of screen s 

: (screen)          ( w h s --- )
  dup scr erase             \ erase screen structure
  tuck s.height w!          \ store width and height in structure
       s.width w! ;

\ ------------------------------------------------------------------------
\ create a named screen structure and allocate buffers

: screen        ( w h --- )
  create                    \ create named screen structure
  here scr allot            \ get address of structure 
  (screen) ;                \ fill in screen structure

\ ------------------------------------------------------------------------
\ close screen - frees buffers, does not close attached windows

: close-screen
  dup s.buffer1 
  dup @ free drop off 
  s.buffer2 
  dup @ free drop off ;
  
\ ========================================================================
