\ window.f      - isforth console windowing words
\ ------------------------------------------------------------------------

 .( loading window.f ) cr

\ ------------------------------------------------------------------------
\ set second attribute for window w (bold, underline etc)

: wprefset ( w n1 --- ) 
  swap w.attrs              \ point to windows second attribute 
  tuck c@ or                \ set attribute bit 
  swap c! ;                 \ store back in window structure

\ ------------------------------------------------------------------------
\ clear second attribute for window w

: wprefclr ( w n1 --- ) 
  not                       \ invert attribute mask for clearing bit
  swap w.attrs              \ point to windows second attribute
  tuck c@ and               \ clear attribute bit
  swap c! ;                 \ store back in window structure

\ ------------------------------------------------------------------------
\ set or clear specific extended attributes

: w>so       ( w --- )   1 wprefset ;   \ standout
: w<so       ( w --- )   1 wprefclr ;
: w>ul       ( w --- )   2 wprefset ;   \ underline
: w<ul       ( w --- )   2 wprefclr ;
: w>rev      ( w --- )   4 wprefset ;   \ reverse video
: w<rev      ( w --- )   4 wprefclr ;
: w>bold     ( w --- )   8 wprefset ;   \ bold
: w<bold     ( w --- )   8 wprefclr ;
: w>alt      ( w --- )  16 wprefset ;   \ alt charset (ibm box charset)
: w<alt      ( w --- )  16 wprefclr ;

\ ------------------------------------------------------------------------
\ set new foreground color for window

: w>fg          ( n1 w --- )
  w.colors                  \ point to window's color byte
  tuck c@ $f0 and           \ mask out old foreground value
  or                        \ or in new value
  swap c! ;                 \ store new attribute

\ ------------------------------------------------------------------------
\ set new background color for window

: w>bg          ( n1 w --- )
  w.colors                  \ point to window's color byte
  tuck c@ $0f and           \ mask out old background value
  swap 4 << or              \ shift new value into position and or in
  swap c! ;                 \ store new attribute

\ ------------------------------------------------------------------------
\ set both fg and bg 

: w>attrib      ( n1 w --- )
  w.colors c! ;

\ ------------------------------------------------------------------------
\ get number of bytes in one window line

<headers

: wlb            ( w --- 4*width )
  w.width w@ 4* ;

\ ------------------------------------------------------------------------
\ calculate window size in bytes

: w-size        ( w --- size )
  dup wlb swap              \ size of one line times
  w.height w@ * ;           \ number of lines in window

\ ------------------------------------------------------------------------
\ set/get window cursor position

: x!    ( x w --- ) w.cx w! ;
: y!    ( y w --- ) w.cy w! ;
: x@    ( w --- x ) w.cx w@ ;
: y@    ( w --- y ) w.cy w@ ;

\ ------------------------------------------------------------------------
\ set position of window in screen

headers>

: winpos    ( x y w --- )
  tuck
  w.yco w!
  w.xco w! ;

\ ------------------------------------------------------------------------
\ is window boxed?

<headers

: ?boxed    ( w --- f )
  w.flags w@ 1 and ;

\ ------------------------------------------------------------------------
\ relocate cursor within window w

headers>

: wat           ( x y w --- )
  dup>r 
  ?boxed dup d+             \ add 1 to x and y if window is boxed
  r> tuck x! y! ;           \ set cursor location in window

\ ------------------------------------------------------------------------
\ erase line n1 of window

\ I really need some better names for some of these words, but there's no
\ way I'm going to call this word "clear-line-n1-of-window" - this is NOT
\ ans forth :/

: elw               ( n1 w --- )
  dup>r w.blank c@ 16 <<    \ blank char for window (usually a bl) 
  r@ w.attrs c@ 8 << or     \ attributes to display chars with
  r@ w.colors c@ or         \ colors to display chars with
  swap                      \ get n1 

  r@ w.width w@ !> ww       \ get and save window width
  r@ wlb *                  \ get n1 * 4* window width 
  r@ w.buffer @ +           \ plus buffer address = address of line n1
  r> ?boxed dup 2* negate   \ if window is boxed...
  +!> ww 4* +               \ subtract 2 from width and add 4 to address

  sp@ 4+ swap               \ point to char and attrib
  ww                        \ for width of window (excluding box)
  for
    2dup 4 cmove            \ copy data into line of window
    4+                      \ advance destination
  nxt
  3drop ;

\ ------------------------------------------------------------------------
\ copy line n1 of window to line n2

: wl-copy           ( n1 n2 w --- )
  dup>r wlb !> ww           \ get width of one line
  r@ w.buffer @ !> wb       \ get buffer address
  ww tuck * -rot *          \ get offset to source and destination lines
  
  r> ?boxed 4* dup>r        \ if window is boxed
  dup d+                    \ add 4 to each line offset (skip box)
  r> 2* negate +!> ww       \ and subtract 8 from width

  wb dup d+ swap            \ convert offsets to addresses
  ww cmove ;

\ ------------------------------------------------------------------------
\ scroll window up

: scroll-up     ( w --- ) 
  dup>r ?boxed              \ 0 or 1 = top line of scroll
  dup 2* negate             \ 0 or 2 lines of box present

  r@ w.height w@ 1- +       \ height -1 minus # box lines
  r> swap                   \ ( line-1 w height --- )
  for
    >r                      \ save window
    1+ dup dup 1-           \ compute source and destination lines
    r@ wl-copy              \ copy line up
    r>
  nxt
  elw ;                     \ erase bottom line of window

\ ------------------------------------------------------------------------

: scroll-dn     ( w --- ) 
  dup>r
  w.height w@ 1-
  r@ ?boxed tuck 2* negate + 
  dup rot + swap r> swap
  for
    >r
    1- dup dup 1+ 
    r@ wl-copy 
    r>
  nxt
  elw ;

\ ------------------------------------------------------------------------

: scroll-lt         ( w --- ) 
  drop ;

\ ------------------------------------------------------------------------

: scroll-rt         ( w --- ) 
  drop ;

\ ------------------------------------------------------------------------
\ move cursor up in window

: cursor-up         ( w --- )
  w.cy 0decr ;

\ ------------------------------------------------------------------------
\ move cursor down in window

: cursor-dn         ( w --- )
  dup y@ 1+ >r              \ increment y coordinate
  dup w.height w@ r> tuck = \ incremented below bottom of window?
  if
    1-                      \ yes - move back into window
    over scroll-up          \ and scroll window up one line
  then
  swap y! ;                 \ store y back in structure

\ ------------------------------------------------------------------------
\ move cursor left in window

: cursor-lt         ( w --- )
  dup x@ ?dup 0=            \ get cursor x and see if its already zero
  if
    dup w.width w@          \ yes make x one past right edge
    over cursor-up          \ and cursor up one line
  then
  1- swap x! ;              \ decrement x and store back in structure

\ ------------------------------------------------------------------------
\ move cursor right in window

: cursor-rt         ( w --- )
  dup x@ 1+ >r              \ get current y coordinate and increment it
  dup w.width w@ r> tuck =  \ get window width - x beyond right edge?
  if
    drop 0                  \ yes - move x back to left edge and
    over cursor-dn          \ move cursor down
  then
  swap x! ;                 \ set new cursor x

\ ------------------------------------------------------------------------
\ write char c to window w without advancing cursor

: (wemit)       ( c w --- )
  dup>r w.buffer @          \ get window buffer address
  r@ x@ 4* +                \ plus x (4 bytes per character cell)
  r@ y@ r@ wlb * +          \ plus 4 * width * y
  swap 16 <<                \ shift char up
  r@ w.attrs c@ 8 << or     \ get attrs for window
  r> w.colors c@ or         \ get colors for window
  swap ! ;

\ ------------------------------------------------------------------------
\ write c to window and advance cursor

: wemit         ( c w --- )
  tuck (wemit)              \ write char
  cursor-rt ;               \ advance cursor

\ ------------------------------------------------------------------------
\ write string a1 of length n1 to window w

: wtype         ( w a1 n1 --- )
  for                       \ for length of string do...
    2dup c@                 \ fetch char
    swap wemit              \ write it to the window
    1+                      \ advance address
  nxt
  2drop ;

\ ------------------------------------------------------------------------
\ write compiled string to specified window

<headers

: (w")          ( w --- )
  sparam wtype ;

\ ------------------------------------------------------------------------
\ compile string to be written to a window

headers>

: w"        ( --- )
  compile (w") ," align ; immediate

\ ------------------------------------------------------------------------

<headers

: .bline    ( rc mc lc w --- )
  tuck wemit
  ww for
    2dup wemit
  nxt  
  nip (wemit) ;

\ ------------------------------------------------------------------------
\ draw top line of box in window w

: top           ( w --- )
  >r 'k' 'q' 'l'
  r@ .bline 
  r> cursor-rt ;

\ ------------------------------------------------------------------------
\ draw bottom line of box in window w

: bottom        ( w --- )
  >r 'j' 'q' 'm'
  r> .bline ;

\ ------------------------------------------------------------------------
\ draw left/right edge character of box

: side          ( w --- w )
  'x' over wemit ;

\ ------------------------------------------------------------------------
\ draw left and right edges of box in window w

: sides         ( w --- w )
  dup w.height w@ 2-        \ for height of box do
  for
    side                    \ write left edge character of box
    ww over w.cx w+!        \ add window width to cursor x
    side                    \ write right edge charater of box
  nxt ;

\ ------------------------------------------------------------------------
\ draw box around edge of window

headers>

: box           ( w --- )
  dup>r 
  w.flags w@ $fffe and
  r@ w.flags w!
  0 dup r@ wat              \ home cursor on window
  r@ w.width w@ 2- !> ww    \ set width of box edges (minus corners)
  r@ w>alt
  r@ top r@ sides bottom    \ draw box
  r@ w.flags dup w@ 
  1 or swap w! 
  r@ w<alt
  0 0 r> wat ;

\ ------------------------------------------------------------------------
\ clear window with default attrib and blank char (doesnt clear box)

: clw         ( w --- )
  dup ?boxed 2* negate
  over w.height w@ +
  for
    dup scroll-up
  nxt 
  drop ;

\ ------------------------------------------------------------------------
\ clear window with checkerboard character

: backfill      ( w --- ) 
  dup w.blank c@ swap
  'a' over w.blank c!
  dup w>alt dup clw  
  dup w<alt
  w.blank c! ;

\ ------------------------------------------------------------------------
\ blit contents of window into its associated screen

<headers

: (.window)    ( w --- )
  dup w.height w@ >r        \ number of lines to draw to screen
  dup w.xco w@ !> wx        \ get x coordinate of window on screen
  dup w.yco w@ !> wy        \ get y coordinate
  dup wlb !> ww             \ get # bytes per window line
  dup w.buffer @ swap       \ get address of window buffer
      w.screen @  
  dup s.width w@ 4* !> sw   \ get screen width
      s.buffer1 @           \ get address of screen buffer
  wy sw * wx 4* + +         \ get addr of upper left char of win in screen 

  r> 
  for
    2dup ww cmove           \ copy 1 line of window into screen buffer
    ww sw d+
  nxt 
  2drop ;

 ' (.window) is .window

\ ------------------------------------------------------------------------
\ allocate buffer for window w

headers>

: walloc        ( w --- f1 )
  dup w-size                \ get number of bytes in window
  allocate                  \ allocate rw/w buffer
  if
    swap w.buffer !         \ store buffer address in window structure
    true
  else
    drop false
  then ;

\ ------------------------------------------------------------------------
\ set width, height and default attributes in window w

: (window)      ( width height w --- )
  dup win erase             \ erase structure      
  tuck w.height w!          \ set window width and height in structure
  tuck w.width w!  

  $07 over w.colors c!
  $00 over w.attrs c!
  bl swap w.blank c! ;

\ ------------------------------------------------------------------------
\ create a named window structure

: window        ( width height --- )
  create here win allot     \ create named window structure
  (window) ;                \ fill in structure entries

\ ------------------------------------------------------------------------
\ close specified window (does not kill the window structure)

: close-win     ( w --- )
  dup w.screen @            \ is this window attached to a screen?
  if
    dup detach              \ if so detach it
  then
  w.buffer dup @ free       \ deallocate window buffer
  drop off ;                \ window no longer has a buffer
 
\ ========================================================================
