
 .( loading see.f ) cr

\ ------------------------------------------------------------------------

 vocabulary debug debug definitions

<headers                    \ everythng except the last word is headerless

\ ------------------------------------------------------------------------

  0 var indent              \ current indent level
  0 var ?indent             \ true = request indent to new line
  0 var eline               \ true if current line is empty

35 var max-width            \ max column when seeing word

\ ------------------------------------------------------------------------
\ conditionally move to new line and indent to current level

: (.indent)
  ?indent eline not and     \ dont move to new line if current 
  if                        \ line is empty
    cr                      \ otherwise move to a new line
    indent 1+ 2* spaces     \ and indent to requested level
    on> eline               \ new line is currently empty
  then
  off> ?indent ;            

\ ------------------------------------------------------------------------

: .indent       ( --- )  on> ?indent (.indent) ; 
: +indent       ( --- )  incr> indent .indent ;
: -indent       ( --- )  decr> indent .indent ;

\ ------------------------------------------------------------------------

\ we have decompiled up to the end of some control structure (if/then etc)
\ and are about to indent to a new line.  if however the xt immediatly
\ following the control structure is the terminating exit then we must not
\ indent or else we will display the semicolon on a line by itself.

: ??indent           ( a1 --- a1 )
  dup @ ['] exit = 
  indent 0= and ?exit
  on> ?indent ;

\ ------------------------------------------------------------------------
\ conditionally indent then write string, emit a char or output a number

: dd            ( ... a1 --- )
  (.indent)                 \ conditionally indent
  execute                   \ execute output function
  off> eline ;              \ current line is no longer empty

\ ------------------------------------------------------------------------
\ type string (indent to new line first if its requested)

: dtype         ( a1 n1 --- )  ['] type dd ;
: demit         ( c1 --- )     ['] emit dd ;
: .d            ( n1 --- )     ['] . dd ;
: >.id          ( a1 --- )     ['] .id dd ;

\ ------------------------------------------------------------------------
\ type a compiled string indenting if needed

: (d")          ( --- )
  sparam dtype ;

\ ------------------------------------------------------------------------
\ compile a decompiler string

: d"            ( --- )
  compile (d") ," align ; immediate

\ ------------------------------------------------------------------------
\ get next xt from a colon definition, advance address

: $@+           ( a1 --- a2 xt )
  dup 4+                    \ advance address
  swap @ ;                  \ fetch xt
\ ------------------------------------------------------------------------

\ ------------------------------------------------------------------------
\ get address that a branch is pointing to.

: b-target      ( a1 --- a2 )
  dup @ $03fffffe and       \ get branch delta
  6 << 6 >>                 \ sign-extend
  + ;                       \ and add it to the instruction's address.

\ ------------------------------------------------------------------------
\ creates words to test what sort of word we are trying to decompile

: ?cfa          ( a1 --- a1 f1 )
  create ' ,                \ parse next word and compile its cfa 
  does>                     \ test if cfa a1 = compiled cfa
    @                       \ fetch addr to check branch target against 
    over b-target           \ get target of branch instr at a1
    = ;                     \ same ?

\ ------------------------------------------------------------------------
\ words to test what sort of word we are trying to decompile

\ v-- this word creates
\      v-- this word - which tests if cfa of word to decompile...
\                 v-- ;uses this word

  ?cfa colon?     nest
  ?cfa variable?  dovariable
  ?cfa constant?  doconstant
  ?cfa defered?   dodefer

\ everything else is assumed to be a coded def

\ ------------------------------------------------------------------------
\ emit a quote char and a space

: quote         ( --- )
  '"' demit $20 demit ;

\ ------------------------------------------------------------------------
\ decompile a ." or an abort" 

\ a1 is the compiled counted string
\ a2 is address of xt following the (.") or (abort")

: (see")        ( a1 --- a2 )
  count 2dup                \ get address and length of string
  quote dtype quote         \ display string wrapped up in quotes
  + aligned ??indent ;

\ ------------------------------------------------------------------------
\ display some simple word types...

: see-defered   ( a1 --- ) ." defer " >.id ;
: see-variable  ( a1 --- ) ." variable " >.id ;
: see-constant  ( a1 --- ) dup >body @ . ." constant " >.id ;

\ ------------------------------------------------------------------------
\ display some things found in a colon definition

: see-."        ( a1 --- a2 ) .indent '.' demit (see") ;
: see-abort"    ( a1 --- a2 ) .indent d" abort" (see") ;
: see-lit       ( a1 --- a2 ) $@+ .d ;

\ -------------------------------------------------------------------------
\ decompile a complete case: statement

: see-case      ( a1 --- a2 )
  .indent                   \ move to new line - indent to current level
  d" case:"                 \ display start of case
  +indent                   \ increment indent, move to new indent level
  $@+ >r                    \ get case exit point (a2)
  $@+ >r                    \ get default case
  $@+                       \ get case option count
  2dup 4* + -rot            \ ( vector-table option-table count --- )
  for
    .indent
    $@+ .d swap d" opt "    \ get and display one case option
    $@+ >.id swap           \ display option vector
  nxt
  2drop 
  r> ?dup                   \ see if there's a default
  if
    d" dflt " >.id          \ show default case
  then
  -indent d" ;case "
  r> ??indent ;

\ ------------------------------------------------------------------------
\ discard next xt from word to decompile

: @+d           ( a1 --- a2 ) $@+ drop ;
: >i@d          ( a1 --- a2 ) +indent @+d ;

\ ------------------------------------------------------------------------
\ decompile if else then

: see-if        ( a1 --- a2 ) .indent d" if" >i@d ;
: see-else      ( a1 --- a2 ) -indent d" else" >i@d ;
: see-then      ( a1 --- a1 ) -indent d" then " ??indent ;

\ ------------------------------------------------------------------------

: see-?:        ( a1 --- a2 )
  .indent d" ?:"
  +indent $@+ .id 
  decr> indent ;

\ ------------------------------------------------------------------------
\ decompile do loops

: see-do        ( a1 --- a2 ) .indent d" do" >i@d ; 
: see-?do       ( a1 --- a2 ) .indent d" ?do" >i@d ;
: see-loop      ( a1 --- a2 ) -indent d" loop " @+d ??indent ;
: see-+loop     ( a1 --- a2 ) -indent d" +loop " @+d ??indent ;

\ ------------------------------------------------------------------------
\ decompile begin while repeat until again 

: see-begin     ( a1 --- a1 ) .indent d" begin" +indent ;
: see-while     ( a1 --- a2 ) -indent d" while" >i@d ;
: see-repeat    ( a1 --- a2 ) -indent d" repeat " @+d ??indent ;
: see-until     ( a1 --- a2 ) -indent d" until " @+d ??indent ;
: see-again     ( a1 --- a2 ) -indent d" again " @+d ??indent ;

\ ------------------------------------------------------------------------
\ decompile for/nxt

: see-for       ( a1 --- a2 ) .indent d" for" +indent ;
: see-nxt       ( a1 --- a2 ) -indent d" nxt " @+d ??indent ;

\ ------------------------------------------------------------------------
\ decompile !> +!> incr> decr> on> off>

: see%          ( a1 --- a2 ) $@+ body> >.id ;

: see-!>        ( a1 --- a2 ) d" !> " see% ;
: see-+!>       ( a1 --- a2 ) d" +!> " see% ;
: see-incr>     ( a1 --- a2 ) d" incr> " see% ;
: see-decr>     ( a1 --- a2 ) d" decr> " see% ;
: see-on>       ( a1 --- a2 ) d" on> " see% ;
: see-off>      ( a1 --- a2 ) d" off> " see% ;

\ ------------------------------------------------------------------------
\ does ;code part point to dodoes?

 : ?does>        ( a1 --- a2 )
  dup @ $95e1fffc =       \ is `stwu t,-4(sp)' ?  
  over 4+ @ $7de802a6 =   \ is `mflr t' ?
  and over 8 +
  b-target ['] nest =     \ is `b nest' ?
  and if
    .indent d" does>"
    .indent
    12 + exit
  then
  ." ;code "                \ this part is wrong but ill fix it when
  r>drop ;                  \ the disassembler is done

  
\ ------------------------------------------------------------------------
\ defered so disassembler can patch into it

 defer see-;code
 ' ?does> is see-;code

\ ------------------------------------------------------------------------

: see-compile   ( a1 --- a2 ) d" compile " $@+ >.id ;

\ ------------------------------------------------------------------------

: ?xt           ( a1 --- a1 )
  create                    \ create word to test if xt to decompile
  ' ,                       \ has this address in its cfa
  does>
    @ over =                \ fetch cfa to test for and compare it
    if                      \ with cfa of word to decompile
      drop r> @ execute     \ execute this words decompiler
      exit                  \ does not return to caller but one up from
    then
    r> 4+ >r ;              \ skip decompile vector and return

\ ------------------------------------------------------------------------

\ v-- this word
\     v-- creates this word which checks if next xt to decompile is 
\              v-- this word
 
  ?xt lit?     (lit)
  ?xt ."?      (.")
  ?xt abort"?  (abort")
  ?xt case?    docase
  ?xt if?      doif
  ?xt else?    doelse
  ?xt then?    dothen
  ?xt ?:?      ?:
  ?xt do?      (do)
  ?xt ?do?     (?do)
  ?xt loop?    (loop)
  ?xt +loop?   (+loop)
  ?xt begin?   dobegin
  ?xt while?   ?while
  ?xt repeat?  dorepeat
  ?xt again?   doagain
  ?xt until?   ?until
  ?xt !>?      %!>
  ?xt +!>?     %+!>
  ?xt incr>?   %incr>
  ?xt decr>?   %decr>
  ?xt on>?     %on>
  ?xt off>?    %off>
  ?xt compile? compile
  ?xt ;code?   ;code
  ?xt for?     dofor
  ?xt nxt?     (nxt)
  
\ ------------------------------------------------------------------------
\ decompile next xt from colon definition being seen

: (see:)     ( a1 xt --- a1 )
  lit?     see-lit          \ if xt is a (lit) decompile a literal
  ."?      see-."           \ if xt is a ." decompile a ."
  abort"?  see-abort"       \ ...
  case?    see-case
  if?      see-if
  else?    see-else
  then?    see-then
  ?:?      see-?:
  do?      see-do
  ?do?     see-?do
  loop?    see-loop
  +loop?   see-+loop
  begin?   see-begin
  while?   see-while
  repeat?  see-repeat
  until?   see-until
  again?   see-again
  !>?      see-!>
  +!>?     see-+!>
  incr>?   see-incr>
  decr>?   see-decr>
  on>?     see-on>
  off>?    see-off>
  compile? see-compile
  ;code?   see-;code
  for?     see-for
  nxt?     see-nxt

\ if any of the above words finds a match we dont get this far

  ['] exit over =           \ don't display an 'exit' unless its not
  indent 0= and             \ the one compiled by the ;
  if
    drop exit
  then
  >.id ;

\ ------------------------------------------------------------------------
\ see if next xt to decompile is an immediate word

: ?immediate        ( a1 --- a1 )
  dup >name                 \ point at nfa
  ?dup 0= ?exit             \ headerless word?
  c@ $40 and                \ no... imediate ?
  0= ?exit
  d" [compile] " ;

\ ------------------------------------------------------------------------
\ decompile a complete colon definition

: see-colon         ( a1 --- )
  off> indent cr            \ reset indent level
  d" : " dup >.id           \ show the : and the word name
  .indent                   \ indent
  >body                     \ point at body of colon definition
  begin
    $@+ dup>r               \ fetch next xt and decompile it
    ?immediate              \ immediate xt's need a [compile]
    (see:)        
    r> ['] exit =           \ if it was an exit and we are not indented
    indent 0= and           \ then this is the end of the definition
  until                     \ otherwise keep going
  d" ;" drop ;              \ show the ;

\ ------------------------------------------------------------------------

: ?see       ( a1 --- | a1 )
  r> tuck @ execute         \ fetch word type test and execute it
  if                        \ word type match?
    swap 4+ @ execute       \ yes, fetch word type decompile function
    exit                    
  then 
  swap 8 + >r ;             \ no match skip ret addr past decomp vector

\ -----------------------------------------------------------------------

: (see)         ( a1 --- a1 )
  dup
  ?see defered?  see-defered 
  ?see variable? see-variable 
  ?see constant? see-constant 
  ?see colon?    see-colon  ;

\ ------------------------------------------------------------------------
\ decompiles almost anything :)

headers>

: see       ( --- )
  max-width !> idw  
  ['] .indent is .idcr
  cr ' (see)               \ decompile word
  >name ?dup 0= ?exit      \ see if decompiled word is immediate
  c@ $40 and
  if
    ."  immediate"
  then 
  cr ;

\ ======================================================================== 
